VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
END
Attribute VB_Name = "AWE_Sheet"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = False
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
'===================================================================================================
' EXCEL CLASS MODULE: AWE_Sheet
'---------------------------------------------------------------------------------------------------
' OVERVIEW:
'  - Sub-second performance for 100K+ rows, robust error handling, and best practices.
'  - Streamlines your code so you can focus on solutions, not syntax.
'  - Every function adapts to structural sheet changes without your code breaking or needing rewrites,
'    including: Dynamic sheets, Protected sheets, Tables, Shared and Network workbooks
'
'     FREE: Download AWE_Sheet at www.AutomateWithExcel.com
'  License: Freeware (personal or commercial use with attribution)
'---------------------------------------------------------------------------------------------------
' FEATURES:
'  - SmartFilter - 50% faster than AutoFilter, optimized accuracy for Excel's maximum sheet size.
'  - SmartCells - Lightning-fast cell retrieval by column name, search strings, wildcards, and more.
'  - Dynamic column mapping and sheet metadata ensure structural changes never break your code.
'  - All functions seamlessly support ListObjects, standard ranges, and protected sheets.
'  - Supports simultaneous open workbooks: Teams, SharePoint, Network, Local.
'  - Robust error handling with call-stack functionality for better code maintenance.
'  - Scales effortlessly from lightweight tasks to enterprise-level workflowswithout rewrites.
'  - Designed to integrate seamlessly with next-gen performance tools from www.AutomationWithExcel.
'---------------------------------------------------------------------------------------------------
' AUTHOR: Mike Libby  30+ years of Excel/VBA expertise, with GPT-4 collaboration.
'===================================================================================================
' VERSION HISTORY:
'  - 2025.03.30 - Initial release.
'===================================================================================================
Option Explicit

'###################################################################################################
' Global Class Variables - Engineered for High-Performance Worksheet Automation
'###################################################################################################

' Core Variables
Private m_Workbook As Workbook          ' Associated workbook
Private m_Worksheet As Worksheet        ' Associated worksheet
Private m_ListObject As ListObject      ' Associated table (if any)

' Column and Row Management
Private m_HdrRowNb As Long              ' Header row number for column-based operations
Private m_ColumnHdrDict As Object       ' Optimized dictionary for rapid column name-to-number mapping

' SmartFilter Optimization
Private m_FindResultsDict As Object     ' Optimized dictionary for high-speed SmartFilter lookups
Private m_FindResultsColl As Collection ' Collection storing filtered row references for instant access
Private m_FindResultsColCtr As Long     ' Tracks number of columns evaluated in SmartFilter
Private m_FindResultsLastRow As Long    ' Stores last used row to avoid unnecessary recalculations
Private m_FindColumnType As VbVarType   ' Column type being evaluated during SmartFilter execution

' DataBodyRangeX Optimization
Private m_EmptyEndRowDetection As Boolean   ' Excludes empty end rows from DataBodyRangeX for efficiency

' RegEx Object(s)
Private m_FindCriteriaOpRegex As Object ' Pre-compiled regex for fast criteria parsing

'###################################################################################################
' Enumerators - Define allowable options for AWE_Sheet operations
'###################################################################################################

' Enum RangeConvertType - Controls SmartCells output, allowing either a Range reference or extracted values
Public Enum RangeConvertType: rctRange = 0: rctValue = 1: End Enum

' Enum AWE_StateType - Standardized status codes for validating workbook, worksheet, and data integrity
Public Enum AWE_StateType
    State_Workbook = 1000       ' Workbook is valid
    State_ReadOnly = 1001       ' Workbook is read-only
    State_Worksheet = 1002      ' Worksheet is valid
    State_HasData = 1003        ' Worksheet contains data
    State_ListObject = 1005     ' ListObject (Table) exists and is valid
    State_HeaderRow = 1006      ' Header row is valid
    State_HeaderColumn = 1007   ' Header column is valid
    State_Protection = 1008     ' Workbook or worksheet is protected
    State_SheetData = 1009      ' Data exists on the sheet
    State_Parameter = 1010      ' Parameter is valid
    State_Search = 1011         ' Search operation is valid
End Enum

'###################################################################################################
' VALIDATION AND ERROR HANDLING - Ensures System Stability and Traceability
'###################################################################################################
' Returns the numeric value of an AWE_StateType enum.
Public Property Get CStateType(EnumName As AWE_StateType) As Long:  CStateType = EnumName: End Property

'---------------------------------------------------------------------------------------------------
' IsValidState - Prevents invalid operations by ensuring required system states are met.
' Params:  stateType (AWE_StateType) - The state to validate.
'          criterion() (Optional) - Additional parameters for state-specific checks.
' Returns: Boolean - True if the state is valid, otherwise False.
'---------------------------------------------------------------------------------------------------
Private Function IsValidState(stateType As AWE_StateType, ParamArray Criterion()) As Boolean
    On Error Resume Next
    Select Case stateType
        Case State_Workbook:     IsValidState = Not m_Workbook Is Nothing: Exit Function
        Case State_ReadOnly:     IsValidState = Not m_Workbook.ReadOnly: Exit Function
        Case State_Worksheet:    IsValidState = Not m_Worksheet Is Nothing: Exit Function
        Case State_HasData:      IsValidState = Not (m_Worksheet.UsedRange.Cells.Count = 1 And _
                                                     IsEmpty(m_Worksheet.UsedRange.Cells(1, 1)))
                                                     Exit Function
        Case State_Protection:   IsValidState = Not (m_Worksheet.ProtectContents Or _
                                                     ThisWorkbook.ProtectStructure)
                                                     Exit Function
        Case State_SheetData:    IsValidState = (m_Worksheet.UsedRange.rows.Count > 1)
        Case State_HeaderRow:    IsValidState = HeaderRowNumber > 0 And _
                                                HeaderRowNumber < m_Worksheet.rows.Count
                                                Exit Function
        Case State_HeaderColumn: IsValidState = m_ColumnHdrDict.Exists(Criterion(0)(0)): Exit Function

        ' Check if ListObject (Table) is valid
        Case State_ListObject
            If m_ListObject Is Nothing Then Exit Function
            If m_ListObject.ListColumns.Count = 0 Then Exit Function
    End Select
    
    IsValidState = True ' If this point is reached, then all checks passed
End Function

'---------------------------------------------------------------------------------------------------
' ValidateState - Enforces critical system states and raises errors when conditions are not met.
' Params:  stateType (AWE_StateType) - The state to validate.
'          criterion() (Optional) - Additional parameters for state-specific checks.
'---------------------------------------------------------------------------------------------------
Private Sub ValidateState(stateType As AWE_StateType, ParamArray Criterion())
    If Not IsValidState(stateType, Criterion) Then
        Select Case stateType
            Case State_Workbook:     RaiseError State_Workbook, "ValidateState", "Workbook is not initialized."
            Case State_ReadOnly:     RaiseError State_ReadOnly, "ValidateState", "Workbook is in Read-Only mode."
            Case State_Worksheet:    RaiseError State_Worksheet, "ValidateState", "Worksheet is not initialized."
            Case State_HasData:      RaiseError State_HasData, "ValidateState", "Worksheet has no data."
            Case State_ListObject:   RaiseError State_ListObject, "ValidateState", "ListObject (Table) is missing or invalid."
            Case State_HeaderRow:    RaiseError State_HeaderRow, "ValidateState", "Headers row is missing or invalid."
            Case State_HeaderColumn: RaiseError State_HeaderColumn, "ValidateState", "Header column name is missing or invalid."
            Case State_Protection:   RaiseError State_Protection, "ValidateState", "Worksheet or workbook is protected."
            Case State_SheetData:    RaiseError State_SheetData, "ValidateState", "Worksheet does not contain structured data."
        End Select
    End If
End Sub

'---------------------------------------------------------------------------------------------------
' RaiseError - Standardized error handling for full traceability and system stability.
' Params:  errorCode (AWE_SheetError) - The error code to raise.
'          functionName (String) - Name of the calling function for error context.
'          description (String) - Detailed error message to display.
'---------------------------------------------------------------------------------------------------
Public Sub RaiseError(errorCode As Long, FunctionName As String, Optional description As String = "")
    Dim fullSource As String, fullDesc As String
    fullSource = "AWE_Sheet::" & FunctionName & IIf(Err.Source <> "", "->" & Err.Source, "")
    fullDesc = FunctionName & ": " & description & IIf(Err.description = "", "", vbLf) & Err.description

    ' Prevent redundant errors if already raised from the same function
    If InStrRev(Err.Source, "::" & FunctionName) > 0 Then _
        Err.Raise Err.Number, Err.Source, Err.description
    Err.Raise errorCode, fullSource, fullDesc
End Sub

'###################################################################################################
' SHEET INITIALIZATION AND COLUMN MAPPING - Automates Setup and Prevents Breakage
'###################################################################################################
'---------------------------------------------------------------------------------------------------
' Class Constructor - Initializes core variables for high-speed operations
'---------------------------------------------------------------------------------------------------
Private Sub Class_Initialize()
    Set m_ColumnHdrDict = CreateObject("Scripting.Dictionary"): m_ColumnHdrDict.CompareMode = vbTextCompare
    Set m_FindCriteriaOpRegex = CreateObject("VBScript.RegExp")
    With m_FindCriteriaOpRegex: .pattern = "^(<>|<=|>=|=<|=>|<|>|=)?(.*)": .Global = False: End With
End Sub

'---------------------------------------------------------------------------------------------------
' Initialize - Automates opening workbook and sheet setup, preventing errors and manual fixes.
' Params:  SheetNameOrObj (Variant) - Worksheet, sheet name, or ListObject.
'          HeaderRowNumber (Optional) - Header row number (default: 1).
'          WorkbookFileNameOrObj (Optional) - Workbook object or file name (default: ThisWorkbook).
'          OpenReadOnly (Optional) - Open workbook as read-only (default: False).
' Example: aweSh.Initialize "Sheet1"         ' Initialize with a local sheet.
'          aweSh.Initialize "RemoteSheet", , "C:\File.xlsx", True  ' Readonly remote sheet.
'          aweSh.Initialize "TableSheet"     ' Initialize with first ListObject found.
' Notes: - Reuses open workbooks to prevent errors when trying to reopen the same workbook name.
'        - Detects and integrates ListObjects if available.
'        - Throws an error if multiple ListObjects exist.
'---------------------------------------------------------------------------------------------------
Public Sub Initialize(SheetNameOrObj As Variant, Optional HeaderRowNumber As Long = 1, _
                      Optional WorkbookFileNameOrObj As Variant = Nothing, _
                      Optional OpenReadOnly As Boolean = False)
    On Error GoTo ErrorHandler

    ' Reset core variables
    m_HdrRowNb = 0: m_ColumnHdrDict.RemoveAll
    Set m_ListObject = Nothing: Set m_Workbook = Nothing: Set m_Worksheet = Nothing
    m_EmptyEndRowDetection = True: SmartFilterClear

    ' Set Workbook
    Select Case TypeName(WorkbookFileNameOrObj)
        Case "Workbook": Set m_Workbook = WorkbookFileNameOrObj
        Case "String"
            If Not IsWorkbookOpen(CStr(WorkbookFileNameOrObj), m_Workbook) Then
                On Error Resume Next
                Set m_Workbook = Workbooks.Open(CStr(WorkbookFileNameOrObj), ReadOnly:=OpenReadOnly)
                On Error GoTo 0
            End If
        Case Else: Set m_Workbook = ThisWorkbook
    End Select
    ValidateState AWE_StateType.State_Workbook

    ' Set Worksheet
    Set m_Worksheet = ResolveWorksheet(SheetNameOrObj)
    ValidateState AWE_StateType.State_Worksheet

    ' Auto-detect ListObject if present
    On Error Resume Next
    If m_Worksheet.ListObjects.Count > 0 Then
        Set m_ListObject = m_Worksheet.ListObjects(1)
        m_HdrRowNb = m_ListObject.HeaderRowRange.Row
    End If
    On Error GoTo 0

    ' Map and Validate Header Row
    m_HdrRowNb = HeaderRowNumber: MapColumnNumbers
    Exit Sub

ErrorHandler: RaiseError Err.Number, "Initialize", Err.description
End Sub

'---------------------------------------------------------------------------------------------------
' MapColumnNumbers - Maps column names to numbers, ensuring integrity and preventing errors.
' Notes: - Clears existing mappings and validates header metadata.
'        - Adds column names to the dictionary, ensuring uniqueness.
'        - Raises an error if duplicate column names are found.
'---------------------------------------------------------------------------------------------------
Public Sub MapColumnNumbers()
    On Error GoTo ErrorHandler
    ValidateState State_HeaderRow
    m_ColumnHdrDict.RemoveAll

    ' Retrieve header values
    Dim colIndex As Long, arr: arr = HeaderRowRangeX.Value
    If Application.WorksheetFunction.CountA(HeaderRowRangeX) = 0 Then Exit Sub
    If Not SafeIsArray(arr) Then If Len(Trim(arr)) > 0 Then m_ColumnHdrDict.Add Trim(arr), 1: Exit Sub

    ' Process multiple headers
    For colIndex = LBound(arr, 2) To UBound(arr, 2)
        Dim colName As String: colName = Trim(arr(1, colIndex))
        If Len(colName) > 0 Then
            If m_ColumnHdrDict.Exists(colName) Then
                RaiseError State_HeaderColumn, "MapColumnNumbers", "Duplicate column: " & colName
            Else
                m_ColumnHdrDict.Add colName, colIndex
            End If
        End If
    Next colIndex
    Exit Sub

ErrorHandler: RaiseError Err.Number, "MapColumnNumbers", "Error mapping columns."
End Sub

'###################################################################################################
' SHEET METADATA AND RANGES - Retrieves Key Worksheet References and Data Boundaries
'###################################################################################################
' ParentWorkbook - Returns the associated workbook, ensuring it is valid.
Public Property Get ParentWorkbook() As Workbook: ValidateState State_Workbook: Set ParentWorkbook = m_Workbook: End Property

' ParentWorksheet - Returns the associated worksheet, ensuring it is valid.
Public Property Get ParentWorksheet() As Worksheet: ValidateState State_Worksheet: Set ParentWorksheet = m_Worksheet: End Property

' ParentTable - Returns the associated table (ListObject), ensuring it is valid.
Public Property Get ParentTable() As ListObject: ValidateState State_ListObject: Set ParentTable = m_ListObject: End Property

' IsTable - Returns True if the worksheet contains a valid Table (ListObject).
Public Property Get IsTable() As Boolean: IsTable = Not m_ListObject Is Nothing: End Property

' IsSheetEmpty - Returns True if the worksheet contains no data.
Public Property Get IsSheetEmpty() As Boolean
    On Error Resume Next: Dim rng As Range: Set rng = m_Worksheet.UsedRange
    If rng.Cells.Count = 1 And IsEmpty(rng.Cells(1, 1)) Then IsSheetEmpty = True _
    Else IsSheetEmpty = Application.CountA(rng) = 0
    On Error GoTo 0
End Property

' IsDataBodyEmpty - Returns True if the worksheet has headers but no data below the header row.
Public Property Get IsDataBodyEmpty() As Boolean
    On Error Resume Next
    Dim dataRange As Range: Set dataRange = DataBodyRangeX(False)
    IsDataBodyEmpty = (dataRange Is Nothing Or dataRange.rows.Count = 1)
    On Error GoTo 0
End Property

' HeaderRowNumber - Returns the stored header row number for column mapping.
Public Property Get HeaderRowNumber() As Long: HeaderRowNumber = m_HdrRowNb: End Property

' HeaderRowRangeX - Returns the range of the header row used for column mapping.
Public Property Get HeaderRowRangeX() As Range
    On Error GoTo ErrorHandler
    Set HeaderRowRangeX = m_Worksheet.Range(m_Worksheet.Cells(m_HdrRowNb, 1), _
                                                  m_Worksheet.Cells(m_HdrRowNb, LastColumnNumber))
    Exit Property
ErrorHandler:
    ValidateState AWE_StateType.State_Worksheet
    RaiseError Err.Number, "HeaderRowRangeX", "Error retrieving header range"
End Property

' RangeX - Returns the full dataset range, combining headers and data.
Public Property Get RangeX() As Range
    On Error GoTo ErrorHandler
    Dim dataRange As Range, area As Range
    Dim lastRow As Long

    ' Retrieve data range
    Set dataRange = DataBodyRangeX(False)

    ' Ensure the data range exists
    If Not dataRange Is Nothing Then
        With m_Worksheet
            Set area = dataRange.Areas(dataRange.Areas.Count)  ' Retrieve the last area
            lastRow = area.rows(area.rows.Count).Row  ' Find the last row
            Set RangeX = .Range(.Cells(m_HdrRowNb, 1), .Cells(lastRow, dataRange.columns.Count))
        End With
    Else
        Set RangeX = HeaderRowRangeX  ' If no data, just return header row
    End If
    Exit Property

ErrorHandler:
    RaiseError Err.Number, "RangeX", "Error retrieving RangeX"
    Set RangeX = Nothing
End Property

'---------------------------------------------------------------------------------------------------
' DataBodyRangeX - Returns the data range below the header row, either from a ListObject (table) or
'                  UsedRange, while excluding trailing empty rows (ghost rows) at the end of the sheet.
' Params:  RaiseSheetEmptyError (Boolean, Optional) - If True, raises an error if the sheet is empty.
' Returns: Range - The entire data range or Nothing if no data rows exist.
' Notes:   - Supports both ListObjects (tables) and standard ranges.
'          - Ensures the range contains the last data row unless Empty End Rows were previously deleted.
'          - If no data exists, DataBodyRangeX will return a single row beyond the header.
'---------------------------------------------------------------------------------------------------
Public Function DataBodyRangeX(Optional RaiseSheetEmptyError As Boolean = False) As Range
    On Error GoTo ErrorHandler
    Dim usedRng As Range, isAutoFilterMode As Boolean, IsFiltered As Boolean, lastArea As Range
    Dim frstRow As Long, lastRow As Long, LastDataRow As Long, lastCol As Long
    
    With m_Worksheet
        ' Determine data range based on ListObject or UsedRange
        If Not m_ListObject Is Nothing Then
            ' Use ListObject's data body range if available
            Set usedRng = m_ListObject.databodyrange
            If usedRng Is Nothing Then
                frstRow = m_HdrRowNb + 1: lastRow = m_HdrRowNb + 1
                Set usedRng = .Range(.Cells(frstRow, 1), .Cells(lastRow, m_ListObject.Range.columns.Count))
            End If
        Else
            ' If no ListObject, use UsedRange to determine the data range
            Set usedRng = .UsedRange: If usedRng.Address = "$A$1" Then Set DataBodyRangeX = usedRng: Exit Function
            frstRow = m_HdrRowNb + 1: lastRow = Application.Max(usedRng.rows(usedRng.rows.Count).Row, frstRow)
            lastCol = Application.Max(usedRng.columns.Count, 1)
            Set usedRng = .Range(.Cells(frstRow, 1), .Cells(lastRow, lastCol))
        End If
        
        ' Verify the last true data row, unless EmptyEndRowDetection was explicitly turned off for performance
        If m_EmptyEndRowDetection Then
            Set lastArea = usedRng.Areas(usedRng.Areas.Count)
            If Not lastArea Is Nothing Then
                 ' Identify the true last row with actual data by scanning backward
                lastRow = lastArea.rows(lastArea.rows.Count).Row
                For LastDataRow = lastRow To m_HdrRowNb + 1 Step -1
                    ' Exit the loop when a row with data is found
                    If Application.CountA(usedRng.rows(LastDataRow - m_HdrRowNb)) > 0 Then Exit For
                Next LastDataRow
                LastDataRow = Application.Max(m_HdrRowNb + 1, LastDataRow)
                ' Adjust the data range to exclude empty ghost rows
                Set usedRng = .Range(.Cells(m_HdrRowNb + 1, 1), .Cells(LastDataRow, m_ColumnHdrDict.Count))
            End If
        End If
    End With

    ' Handle Empty End Rows (if enabled) to ensure only actual data is included in the range
    If RaiseSheetEmptyError And usedRng Is Nothing Then _
        RaiseError AWE_StateType.State_SheetData, "DataBodyRangeX", "Unable to retrieve range."

    Set DataBodyRangeX = usedRng
    Exit Function

ErrorHandler:
    RaiseError Err.Number, "DataBodyRangeX", "Error retrieving data body range"
    Set DataBodyRangeX = Nothing
End Function

' EmptyEndRowDetection (Get) - Returns whether empty trailing rows are excluded from the data range.
Public Property Get EmptyEndRowDetection() As Boolean: EmptyEndRowDetection = m_EmptyEndRowDetection: End Property

' EmptyEndRowDetection (Let) - Enables or disables detection of empty trailing rows.
Public Property Let EmptyEndRowDetection(ByVal Value As Boolean): m_EmptyEndRowDetection = Value: End Property

'---------------------------------------------------------------------------------------------------
' DeleteEmptyEndRows - Removes empty trailing rows (ghost rows) to ensure UsedRange reflects actual data.
' Returns: Boolean - True if Empty End Rows were successfully removed, False otherwise.
' Notes:   - Updates UsedRange to ensure it contains the absolute last row of data.
'          - Turn off Empty End Rows detection: EmptyEndRowDetection = Not DeleteEmptyEndRows.
'---------------------------------------------------------------------------------------------------
Public Function DeleteEmptyEndRows() As Boolean
    Dim lastRow As Long, LastDataRow As Long, lastArea As Range, tmpRng As Range, usedRng As Range
    Dim isScreenUpdating As Boolean, prevEmptyEndRowDetection As Boolean
    
    ' Ensure the function does not attempt modifications on protected sheets
    On Error Resume Next
    If m_Worksheet.AutoFilterMode Then m_Worksheet.ShowAllData
    On Error GoTo ErrorHandler
    
    ' Store previous Empty End Row detection setting and disable it for this operation
    prevEmptyEndRowDetection = EmptyEndRowDetection: EmptyEndRowDetection = False
    
     ' Retrieve the current data range. Exit function if empty
    Set usedRng = DataBodyRangeX: If usedRng Is Nothing Then Exit Function

    With usedRng.Worksheet
        ' Iterate backwards to find the actual last row with data
        Set lastArea = usedRng.Areas(usedRng.Areas.Count)
        If Not lastArea Is Nothing Then
            ' Identify last row with data
            lastRow = lastArea.rows(lastArea.rows.Count).Row
            For LastDataRow = lastRow To m_HdrRowNb + 1 Step -1
                ' Exit the for loop if a row with actual data is found
                If Application.CountA(usedRng.rows(LastDataRow - m_HdrRowNb)) > 0 Then Exit For
            Next LastDataRow
            LastDataRow = Application.Max(m_HdrRowNb + 1, LastDataRow)

            ' Adjust UsedRange to remove Empty End Rows
            Set usedRng = .Range(.Cells(m_HdrRowNb + 1, 1), .Cells(LastDataRow, m_ColumnHdrDict.Count))

            ' If the UsedRange contains ghosts, then delete Empty End Rows and force recalculation
            If LastDataRow <> lastRow Then
                ' Delete empty rows
                On Error Resume Next
                .Range(.Cells(LastDataRow + 1, 1), .Cells(lastRow, 1)).EntireRow.Delete
                On Error GoTo 0

                ' Force recalculation
                isScreenUpdating = Application.ScreenUpdating
                Application.ScreenUpdating = False
                .UsedRange.Calculate
                DoEvents
                
                ' Refrsh the UsedRange after deletion and recalculation
                Set tmpRng = .UsedRange:  Set lastArea = tmpRng.Areas(tmpRng.Areas.Count)
                lastRow = lastArea.rows(lastArea.rows.Count).Row
                Application.ScreenUpdating = isScreenUpdating
            End If
        End If
    End With

    ' Confirm deletion success
    DeleteEmptyEndRows = (LastDataRow = lastRow)
    EmptyEndRowDetection = prevEmptyEndRowDetection
    Exit Function

ErrorHandler: RaiseError Err.Number, "DeleteEmptyEndRows", "Error deleting Empty End Rows: " & Err.description
End Function

' LastRowNumber - Returns the last row containing data, excluding trailing empty rows.
Public Property Get LastRowNumber() As Long
    On Error GoTo ErrorHandler
    Dim rowRange As Range, lastRow As Long
    
    Set rowRange = DataBodyRangeX(False) ' Get the last row from the last area of DataBodyRangeX()
    If Not rowRange Is Nothing Then
        Dim lastArea As Range: Set lastArea = rowRange.Areas(rowRange.Areas.Count)
        lastRow = lastArea.rows(lastArea.rows.Count).Row
    End If
    LastRowNumber = Application.Max(lastRow, m_HdrRowNb) ' Ensure lastRow is never less than the header row
    Exit Property
ErrorHandler: LastRowNumber = m_HdrRowNb: RaiseError Err.Number, "LastRowNumber", "Error retrieving last row number"
End Property

' LastColumnNumber - Returns the last populated column in the worksheet, excluding empty columns.
Public Property Get LastColumnNumber() As Long
    On Error GoTo ErrorHandler
    Select Case True
        Case Not IsValidState(State_HasData): LastColumnNumber = 1
        Case IsValidState(State_ListObject):  LastColumnNumber = m_ListObject.Range.columns.Count
        Case Else: LastColumnNumber = m_Worksheet.Cells(m_HdrRowNb, _
                        m_Worksheet.columns.Count).End(xlToLeft).Column
    End Select
    Exit Property
ErrorHandler: RaiseError Err.Number, "LastColumnNumber", "Error retrieving last column number"
End Property

' RowCount - Returns the number of rows containing data, excluding trailing empty rows.
Public Property Get rowCount() As Long
    On Error GoTo ErrorHandler
    Dim lastArea As Range, dataRange As Range: Set dataRange = DataBodyRangeX()

    ' Determine row count
    If Not dataRange Is Nothing Then
        Set lastArea = dataRange.Areas(dataRange.Areas.Count)
        rowCount = lastArea.rows(lastArea.rows.Count).Row - m_HdrRowNb
    Else
        rowCount = 0
    End If
    Exit Property
ErrorHandler: RaiseError Err.Number, "RowCount", "Error retrieving row count"
End Property

'###################################################################################################
' COLUMN METADATA AND RANGES
'###################################################################################################
' ColumnNames - Returns an array of column names (0-based index).
Public Property Get ColumnNames() As Variant: ColumnNames = m_ColumnHdrDict.keys(): End Property

'---------------------------------------------------------------------------------------------------
' GetColumnNumber - Resolves a column number from a header name (String), index (Long), or Range.
' Params:  ColIdxOrName (Variant) - Column identifier (String, Long, or Range).
' Returns: Long - The resolved column number.
' Notes:   - Raises an error for invalid column types or missing column names.
'---------------------------------------------------------------------------------------------------
Public Function GetColumnNumber(ColIdxOrName As Variant) As Long
    On Error GoTo ErrorHandler
    Dim colType As String: colType = TypeName(ColIdxOrName)
    If colType = "String" And IsNumeric(ColIdxOrName) Then ColIdxOrName = CLng(ColIdxOrName): colType = "Long"

    ' Resolve column number based on type
    Select Case colType
        Case "Long", "Integer": GetColumnNumber = ColIdxOrName ' Directly return index
        Case "String", "Date":  ValidateState State_HeaderColumn, ColIdxOrName
                                GetColumnNumber = m_ColumnHdrDict(ColIdxOrName) ' Lookup column name
        Case "Range":           GetColumnNumber = ColIdxOrName.Column ' Resolve column from Range
        Case Else:              RaiseError State_HeaderColumn, "GetColumnNumber", _
                                    "Invalid column parameter variant. Must be String, " & _
                                    "Date, Long, or Range. Provided: " & TypeName(ColIdxOrName)
    End Select
    
    ' Handle missing column case
    If GetColumnNumber = 0 Then _
       RaiseError State_HeaderColumn, "GetColumnNumber", "Column not found: " & ColIdxOrName
    Exit Function

ErrorHandler: RaiseError Err.Number, "GetColumnNumber", "Error resolving column: " & ColIdxOrName
End Function

'---------------------------------------------------------------------------------------------------
' GetColumnNumber - Resolves a column number from a header name (String), index (Long), or Range.
' Params:  ColIdxOrName (Variant) - Column identifier (String, Long, or Range).
' Returns: String - The resolved column name.
' Notes:   - Raises an error for invalid column types or missing column names.
'---------------------------------------------------------------------------------------------------
Public Function GetColumnName(ColIdxOrName As Variant) As String
    On Error GoTo ErrorHandler
    GetColumnName = m_ColumnHdrDict.keys()(GetColumnNumber(ColIdxOrName) - 1)
    Exit Function
ErrorHandler: RaiseError State_Parameter, "GetColumnName", "Invalid column reference type: " & TypeName(ColIdxOrName)
End Function

'---------------------------------------------------------------------------------------------------
' ColumnsX - Combines one or more specified columns into a single range, supporting both contiguous
'            and non-contiguous column selections. Rows within the range are always contiguous.
' Params:  ColIdxOrName (ParamArray) - List of column indexes (Long) or names (Strings) to combine.
' Returns: Range - A single combined range for all specified columns.
' Notes:   - Raises an error if any column name is invalid or the range cannot be retrieved.
'---------------------------------------------------------------------------------------------------
Public Function ColumnsX(ParamArray ColIdxOrName() As Variant) As Range
    Dim combinedRange As Range, columnRange As Range, databodyrange As Range
    Dim col As Variant, colIndex As Long
    On Error GoTo ErrorHandler

    ' Combine the specified columns into a single range, ensuring that non-contiguous columns are
    ' handled by using Union to merge them into a single range for processing.
    Set databodyrange = DataBodyRangeX
    For Each col In ColIdxOrName
        colIndex = GetColumnNumber(col)
        If combinedRange Is Nothing Then
            Set combinedRange = Intersect(databodyrange, databodyrange.Parent.columns(colIndex))
        Else
            Set combinedRange = Union(combinedRange, Intersect(databodyrange, databodyrange.Parent.columns(colIndex)))
        End If
    Next col

    ' Validate combined range
    If combinedRange Is Nothing Then _
        RaiseError AWE_StateType.State_SheetData, "ColumnsX", "Could not retrieve column range."

    Set ColumnsX = combinedRange
    Exit Function

ErrorHandler: RaiseError Err.Number, "ColumnsX", "Error getting column: " & col
End Function

'###################################################################################################
' SMART SUITE - Reduce code clutter and increase performance using Excel-like functions with 100K+
'               row sub-second performance and 1M+ row accuracy
'               - WITOUT PROTECTED SHEET AND ROW VISIBILITY CONTRAINTS -
' --------------------------------------------------------------------------------------------------
'   SMARTFILTER - AutoFilter-like criteria but 50% faster with accuracy on 1M+ Rows
'   SMARTCELLS  - Multi-column data retrieval by row and column numbers with advanced search criteria.
'###################################################################################################
'###################################################################################################
' SMARTFILTER - Sub-second filtering on 100K+ rows; same search criteria as AutoFilter but 50% faster
'               with unmatched 1M+ row reliability. Supports protected sheets, non-contiguous columns,
'               and ignores AutoFilter row visibility.
'###################################################################################################
'---------------------------------------------------------------------------------------------------
' SmartFilter - Applies filtering logic to a collection of row numbers, progressively narrowing
'               results with each call. Though similar to AutoFilter, it does not modify
'               the worksheets visible display.
' Params:   srchCol (Variant)   - Column name or index to search in.
'           criteria1 (Variant) - Primary filter criteria.
'                               - Can be a single value for exact matching.
'                               - Can be an array, evaluated using OR logic across all elements.
'                               - If an array is provided, criteria2 must be empty.
'           criteria2 (Optional, Variant) - Secondary filter criteria (default: "").
'                                         - Used only when criteria1 is a single value.
'           criteriaOperator (Optional, XlAutoFilterOperator) - Defines how criteria1 and criteria2
'                                                               are combined (default: xlAnd).
'                                                             - xlAnd: Both criteria must be met.
'                                                             - xlOr: Either criteria can be met.
'           RaiseSearchError (Optional, Boolean) - If True, raises an error when no matches are found
'                                                 (useful for debugging unexpected filtering results).
' Returns:  None (results are stored in m_FindResultsColl and retrieved via SmartFilterRows).
' Notes:    - High-speed filtering without protected sheets or row visibility constraints.
'           - Multiple calls refine results using AND logic between columns.
'           - Supports >, <, >=, <=, =, <> operators (default "=") and wildcards (*, ?).
'           - Criteria1 Array use OR logic; if an Array is used on criteria1, criteria2 must be empty.
'           - Use SmartFilterRows to retrieve results and SmartFilterClear to reset.
'           - RaiseSearchError=True triggers an error if no rows match.
'---------------------------------------------------------------------------------------------------
Public Sub SmartFilter(srchCol As Variant, criteria1 As Variant, _
                       Optional criteria2 As Variant = "", _
                       Optional criteriaOperator As XlAutoFilterOperator = xlAnd, _
                       Optional RaiseSearchError As Boolean = True)
    On Error GoTo ErrorHandler

    Dim colRange As Range, rowNbr As Variant, iRow As Long, iCol As Long, arr As Variant
    Dim SrchVal1 As Variant, Op1 As String, SrchVal2 As Variant, Op2 As String

    ' Validate inputs
    If srchCol = "" Then RaiseError State_Parameter, "SmartFilter", "Search column was not defined."
    If criteriaOperator <> xlAnd And criteriaOperator <> xlOr Then _
        RaiseError State_Parameter, "SmartFilter", "Invalid operator: " & criteriaOperator
    If IsNull(criteria1) Or IsNull(criteria2) Then _
        RaiseError State_Parameter, "SmartFilter", "Criteria cannot be null."

    ' Track the number of columns processed in this filter sequence
    ' This ensures that multiple column filters are applied progressively
    m_FindResultsColCtr = m_FindResultsColCtr + 1 ' Increment column counter
    
    ' Retrieve the column range and store values in an array for faster lookups
    If m_FindResultsLastRow = 0 Then
        Set colRange = ColumnsX(srchCol)
        If colRange.rows.Count > 0 Then m_FindResultsLastRow = colRange.rows(colRange.rows.Count).Row
    Else
        With m_Worksheet
            iCol = GetColumnNumber(srchCol)
            Set colRange = .Range(.Cells(m_HdrRowNb + 1, iCol), .Cells(m_FindResultsLastRow, iCol))
        End With
    End If
    m_FindColumnType = VarType(colRange.Cells(1, 1).Value)
    If m_FindColumnType = vbEmpty Then RaiseError State_SheetData, "SmartFilter", "SrchCol is empty."
    arr = colRange.Value
    
    ' Extract search values and operators from criteria1 and criteria2
    ' Wildcards (*, ?) default to "=" if no explicit operator is provided
    If SafeIsArray(criteria1) Then
        If Not SafeIsEmpty(criteria2) Then _
           RaiseError State_Parameter, "SmartFilter", "if Criteria1 is an array then Criteria2 must be empty."
        SrchVal1 = criteria1
    Else
        GetCriteriaParts criteria1, SrchVal1, Op1
        If criteria1 <> "" And Not IsInArray(Op1, Array("<>", "<=", ">=", "<", ">", "=")) Then _
            RaiseError State_Parameter, "SmartFilter", "Invalid criteria1 Operator: " & Op1
        If criteria2 <> "" Then
            GetCriteriaParts criteria2, SrchVal2, Op2
            If Not IsInArray(Op2, Array("<>", "<=", ">=", "<", ">", "=")) Then _
                RaiseError State_Parameter, "SmartFilter", "Invalid criteria2 Operator: " & Op2
        End If
        ' Convert values to correct type
        SrchVal1 = TypeConvert(VarType(colRange.Cells(1, 1).Value), SrchVal1)
        If criteria2 <> "" Then SrchVal2 = TypeConvert(VarType(SrchVal1), SrchVal2)
    End If
    
    ' Apply filtering: Each call further refines the stored row collection
    ' by removing rows that do not match the current column filter
    Set m_FindResultsColl = SmartFilterEvalRows(m_FindResultsColl, arr, SrchVal1, Op1, SrchVal2, Op2, criteriaOperator)
    
    ' If no matches are found and RaiseSearchError is enabled then raise an error
    If m_FindResultsColl.Count = 0 And RaiseSearchError Then _
        RaiseError AWE_StateType.State_SheetData, "SmartFilter", "No matching rows found."
    Exit Sub

ErrorHandler: RaiseError Err.Number, "SmartFilter", Err.description
End Sub

'---------------------------------------------------------------------------------------------------
' SmartFilterEvalRows - Progressively filters rows across multiple calls by evaluating each cell
'                       in a single column against criteria1 and criteria2. With each successive call,
'                       SmartFilterEvalRows only evaluates criteria against rows stored in InputRowNbrs.
'                       Matching rows are transferred to outputRowNbrs. Once all rows are processed,
'                       InputRowNbrs is discarded, and outputRowNbrs becomes the new filtered collection.
'                       This method of progressively reducing the number of rows evaluated with each
'                       call significantly improves performance.
' Params:   InputRowNbrs (Collection) - Collection of rows from a previous filter pass
'                                     - If empty, this is the first column being filtered
'                                     - If not empty, only rows in this collection will be checked
'           ColumnValueArr (Variant)  - Array of column values corresponding to the filter column
'           SrchVal1 (Variant)        - First search value (single value, array, or dictionary for OR logic)
'           Op1 (String)              - Operator for SrchVal1 (>, <, >=, <=, =, <>)
'           SrchVal2 (Variant, Optional) - Second search value (ignored if criteriaOperator = xlOr)
'           Op2 (String, Optional)    - Operator for SrchVal2
'           criteriaOperator (XlAutoFilterOperator, Optional) - xlAnd/xlOr (default: xlAnd)
' Returns:  Collection - A refined collection of row numbers that match the filter criteria
' Notes:    - Uses two collections:
'             1) InputRowNbrs (previous search results) and 2) outputRowNbrs (new refined results).
'           - Each row in InputRowNbrs is evaluated against the new column's criteria.
'           - Matching rows are added to outputRowNbrs, while non-matching rows are discarded.
'           - Once all rows are processed, InputRowNbrs is discarded, and outputRowNbrs becomes
'             the active set.
'           - If SrchVal1 is an array, it is converted to a dictionary to optimize OR-based lookups.
'---------------------------------------------------------------------------------------------------
Private Function SmartFilterEvalRows(InputRowNbrs As Collection, ColumnValueArr As Variant, _
                                    SrchVal1 As Variant, Op1 As String, _
                                    Optional SrchVal2 As Variant = "", Optional Op2 As String = "", _
                                    Optional criteriaOperator As XlAutoFilterOperator = xlAnd) As Collection
    On Error GoTo ErrorHandler
    Dim outputRowNbrs As New Collection, criteriaDict As Object, val As Variant
    Dim iRow As Long, rowNbr As Variant, cellValue As Variant
    
    ' Convert SrchVal1 to a dictionary if it is an array
    ' - This allows for faster lookups when filtering for multiple values (OR logic).
    If SafeIsArray(SrchVal1) Then
        Set criteriaDict = CreateObject("Scripting.Dictionary"): criteriaDict.CompareMode = vbTextCompare
        For Each val In SrchVal1: criteriaDict(val) = "": Next val
        Set SrchVal1 = criteriaDict
    End If
     
    ' Determine if this is the first column being filtered (m_FindResultsColCtr = 1)
    ' - If true, all rows are checked against criteria
    ' - If false, only rows in InputRowNbrs are checked (progressive filtering)
    If m_FindResultsColCtr = 1 Then
        ' First column search - process all rows in ColumnValueArr
        For iRow = LBound(ColumnValueArr, 1) To UBound(ColumnValueArr, 1)
            rowNbr = iRow + m_HdrRowNb ' Adjust to handle row header
            cellValue = ColumnValueArr(iRow, 1)
            
            ' Check if the row matches the criteria
            If EvalCriterias(cellValue, SrchVal1, Op1, SrchVal2, Op2, criteriaOperator) Then
                outputRowNbrs.Add rowNbr
            End If
        Next iRow
    Else
        ' Subsequent column filtering - refine previous results.
        ' - Instead of adding new matches, rows that do not match are removed.
        For Each rowNbr In InputRowNbrs
            iRow = rowNbr - m_HdrRowNb ' Adjust back to match correct indexing
            cellValue = ColumnValueArr(iRow, 1)
            
            ' Keep only rows that match the current filter criteria.
            ' - Filter is being rebuilt here
            If EvalCriterias(cellValue, SrchVal1, Op1, SrchVal2, Op2, criteriaOperator) Then
                outputRowNbrs.Add rowNbr
            End If
        Next rowNbr
    End If
    
    ' Return the filtered collection
    Set SmartFilterEvalRows = outputRowNbrs
    Exit Function

ErrorHandler: RaiseError Err.Number, "SmartFilterEvalRows", Err.description
End Function

'---------------------------------------------------------------------------------------------------
' EvalCriterias - Evaluates a cell value against one or two search criteria, applying AND or OR logic.
'                 If criteria1 is an array, OR logic is applied across all elements in the array.
'                 If criteria1 and criteria2 are both provided, they are evaluated based on
'                 criteriaOperator.
' Params:  CellValue (Variant) - Cell value to compare.
'          SrchVal1 (Variant) - First search value or Dictionary for lookup.
'          Op1 (String) - Operator for SrchVal1.
'          SrchVal2 (Variant, Optional) - Second search value (default: "").
'          Op2 (String, Optional) - Operator for SrchVal2 (default: "").
'          criteriaOperator (XlAutoFilterOperator, Optional) - xlAnd/xlOr (default: xlAnd).
' Returns: Boolean - True if the value matches criteria.
' Notes:   - Supports wildcard filtering (*, ?) for partial matches.
'          - If SrchVal1 is a Dictionary (due to criteria1 being an array), it enables fast lookups
'            for OR-based evaluations.
'          - Ensures robust comparisons by applying operators correctly based on data type.
'---------------------------------------------------------------------------------------------------
Private Function EvalCriterias(cellValue As Variant, SrchVal1 As Variant, Op1 As String, _
                               Optional SrchVal2 As Variant = "", Optional Op2 As String = "", _
                               Optional criteriaOperator As XlAutoFilterOperator = xlAnd) As Boolean
    ' If SrchVal1 is a Dictionary, it means criteria1 was an array
    ' - This enables fast lookups where OR logic is applied to multiple possible values
    If TypeName(SrchVal1) = "Dictionary" Then
        If SrchVal1.Exists(cellValue) Then EvalCriterias = True: Exit Function
    End If
     
    ' Convert CellValue to uppercase if the search column is text-based
    If m_FindColumnType = vbString Then cellValue = UCase(cellValue)
        
    ' If only one criteria is provided, evaluate against SrchVal1 only
    If SrchVal2 = "" Then
        ' Check if CellValue matches SrchVal1 using the specified operator
        If EvalCriterion(cellValue, SrchVal1, Op1) Then EvalCriterias = True: Exit Function
    Else
        ' Apply the correct logic based on criteriaOperator (xlAnd or xlOr)
        If criteriaOperator = xlOr Then
            ' Match if either SrchVal1 OR SrchVal2 criteria is met
            If EvalCriterion(cellValue, SrchVal1, Op1) Then EvalCriterias = True: Exit Function
            If EvalCriterion(cellValue, SrchVal2, Op2) Then EvalCriterias = True: Exit Function
        Else
            ' Match only if BOTH SrchVal1 AND SrchVal2 criteria are met
            If EvalCriterion(cellValue, SrchVal1, Op1) And EvalCriterion(cellValue, SrchVal2, Op2) Then _
               EvalCriterias = True: Exit Function
        End If
    End If
    EvalCriterias = False
End Function

'---------------------------------------------------------------------------------------------------
' EvalCriterion - Evaluates a single value against a search value using the specified operator.
'                 Supports numeric, text, and wildcard-based comparisons.
' Params:  Value (Variant) - The value to compare.
'          SrchVal (Variant) - The search value to evaluate against.
'          Op (String) - The comparison operator (=, <>, <, <=, >, >=, *).
' Returns: Boolean - True if the comparison evaluates as True, otherwise False.
' Notes:   - Ensures proper data type comparisons before evaluating conditions.
'          - Uses "Like" for wildcard comparisons when Op is "*".
'          - Returns False if an unsupported operator is provided.
'---------------------------------------------------------------------------------------------------
Private Function EvalCriterion(Value As Variant, SrchVal As Variant, Op As String) As Boolean
    Select Case Op ' Perform the comparison
        Case ">":   EvalCriterion = (Value > SrchVal)
        Case "<":   EvalCriterion = (Value < SrchVal)
        Case ">=":  EvalCriterion = (Value >= SrchVal)
        Case "<=":  EvalCriterion = (Value <= SrchVal)
        Case "=":   EvalCriterion = (Value = SrchVal)
        Case "<>":  EvalCriterion = (Value <> SrchVal)
        Case "*":   EvalCriterion = (Value Like SrchVal) ' Use "Like" for wildcard matching
        Case Else:  EvalCriterion = False                ' Handle unsupported operators gracefully
    End Select
End Function

'---------------------------------------------------------------------------------------------------
' SmartFilterClear - Resets SmartFilter by clearing all stored row results and tracking variables.
'---------------------------------------------------------------------------------------------------
Public Sub SmartFilterClear()
    On Error Resume Next: m_FindResultsColCtr = 0: m_FindResultsLastRow = 0: On Error GoTo 0
End Sub

'---------------------------------------------------------------------------------------------------
' IsSmartFilterActive - Returns True if SmartFilter has active results stored in m_FindResultsColl.
'                       - Returns False if no filtered rows are currently stored.
'                       - Does not check if SmartFilter was previously calledonly if results exist.
'---------------------------------------------------------------------------------------------------
Public Property Get IsSmartFilterActive() As Boolean: IsSmartFilterActive = (m_FindResultsColl.Count > 0): End Property

'---------------------------------------------------------------------------------------------------
' SmartFilterRows - Retrieves filtered row numbers or row ranges from SmartFilter results.
'                   - By default, returns a collection of row numbers for efficient processing.
'                   - If returnRowRanges = True, returns row ranges including all columns.
'                   - Use row numbers when performing additional logic or calculations.
'                   - Use row ranges when copying or formatting filtered results.
' Params:   returnRowRanges (Boolean, Optional) - If True, returns row ranges instead of row numbers.
' Returns:  Collection - A collection of row numbers or row ranges.
'---------------------------------------------------------------------------------------------------
Public Function SmartFilterRows(Optional returnRowRanges As Boolean = False) As Collection
    On Error GoTo ErrorHandler
    Dim rowNbr As Variant, rngColl As New Collection, lastCol
    
    ' Exit early if no active SmartFilter results
    ' - If no filtering has been applied, return an empty collection
    If m_FindResultsColCtr = 0 Then Set SmartFilterRows = New Collection: Exit Function
    
    ' If returnRowRanges is False, return stored row numbers from m_FindResultsColl
    If Not returnRowRanges Then Set SmartFilterRows = m_FindResultsColl: Exit Function
    
    ' Return row numbers that have been converted into full row ranges
    ' - Retrieves the last column index from the column header dictionary
    lastCol = m_ColumnHdrDict.Count
    For Each rowNbr In m_FindResultsColl
        rngColl.Add m_Worksheet.Range(m_Worksheet.Cells(rowNbr, 1), _
                                        m_Worksheet.Cells(rowNbr, lastCol))
    Next rowNbr
    Set SmartFilterRows = rngColl
    Exit Function
    
ErrorHandler: RaiseError Err.Number, "SmartFilterRows", Err.description
End Function

'---------------------------------------------------------------------------------------------------
' GetCriteriaParts - Extracts the operator and search value from criteria.
'                  - If no operator is provided, "=" is assumed.
' Params:  criteria (Variant) - Criteria string (e.g., ">75", "Task*").
'          SrchVal (String, ByRef) - Outputs the main search value.
'          Op (String, ByRef) - Outputs the operator (defaults to "=" if missing).
' Returns: Boolean - True if SrchVal is not empty, False otherwise.
' Notes:   - Supports operators: =, <>, <, <=, >, >=, =>, =<.
'          - Defaults Op to "=" if no explicit operator is found in criteria.
'          - Detects wildcard characters (*, ?) and sets Op to "*" if present, overriding other operators.
'          - Uses regular expressions to extract criteria parts efficiently.
'---------------------------------------------------------------------------------------------------
Private Function GetCriteriaParts(criteria As Variant, ByRef SrchVal As Variant, ByRef Op As String) As Boolean
    Dim match As Object
    SrchVal = "": Op = "" ' Reset outputs and create RegEx Object if not provided
    
    ' Use regex (created during te class constructor) to extract operator and search value
    Set match = m_FindCriteriaOpRegex.Execute(criteria)
    If Not match Is Nothing Then
        If match.Count > 0 Then
            ' Extract operator if found, otherwise default to "="
            Op = match(0).SubMatches(0): If Op = "" Then Op = "="
            ' Extract search value
            SrchVal = match(0).SubMatches(1)
            ' If criteria contains * or ?, override operator with "*" indicating wildcard match
            If InStr(criteria, "*") > 0 Or InStr(criteria, "?") Then Op = "*"
        End If
    End If
    
    ' Return True if a valid search value was found
    GetCriteriaParts = (SrchVal <> "")
End Function

'---------------------------------------------------------------------------------------------------
' TypeConvert - Converts a value to the expected data type for consistent filtering behavior.
'               - Significantly improves performance by converting the search value type to the
'                 column type before rows are evaluated.
'               - Enforces strict type conversion based on VbVarType.
'               - Strings are converted to uppercase and truncated to 255 characters if necessary.
'               - Raises an error if conversion fails due to incompatible data types.
' Params:  typeData (VbVarType) - The expected data type.
'          convertData (Variant) - The value to be converted.
'          columnName (Optional) - Column name for error messages (default: "").
' Returns: Variant - The converted value, or Null if conversion fails.
'---------------------------------------------------------------------------------------------------
Private Function TypeConvert(typeData As VbVarType, convertData As Variant, _
                             Optional columnName As String = "") As Variant
    On Error GoTo ErrorHandler

    ' Attempt conversion based on expected type
    Select Case typeData
        Case vbDate:    TypeConvert = CDate(convertData)
        Case vbString:  If Len(convertData) > 255 Then RaiseError State_Parameter, "TypeConvert", "Criteria greater than 255."
                        TypeConvert = UCase(CStr(convertData))
        Case vbDouble, vbSingle, vbInteger, vbLong, vbCurrency: TypeConvert = CDbl(convertData)
        Case vbBoolean: TypeConvert = CBool(convertData)
        Case vbEmpty:   TypeConvert = convertData
        Case Else:      Err.Raise State_Parameter, "TypeConvert", "Unexpected Column type."
    End Select
    Exit Function

ErrorHandler:
    ' Append column name if provided
    If columnName <> "" Then columnName = ". Column: " & columnName
    ' Determine error type based on expected data type
    Select Case typeData
        Case vbDate:    RaiseError State_Parameter, "TypeConvert", "Invalid date Format:" & columnName
        Case vbString:  RaiseError State_Parameter, "TypeConvert", "Invalid string Format:" & columnName
        Case vbDouble, vbSingle, vbInteger, vbLong: RaiseError State_Parameter, "TypeConvert", "Invalid numeric Format:" & columnName
        Case vbBoolean: RaiseError State_Parameter, "TypeConvert", "Invalid boolean Format:" & columnName
        Case Else:      RaiseError Err.Number, "TypeConvert", Err.description & columnName
    End Select
    TypeConvert = Null ' Return Null on error
End Function

'###################################################################################################
' SMARTCELLS - SmartCells extends Cells functionality with 100K+ sub-second performance and flexible
'              row-level access using row numbers, search criteria (including wildcards), column
'              names, or ranges. Reduces variable clutter by retrieving row data from multi-columns
'              in a dictionary. Supports protected sheets and multi-column extraction in one call.
'###################################################################################################
'---------------------------------------------------------------------------------------------------
' SmartCells - High-performance row retrieval and lookup function for structured data.
'          Supports direct row access, string-based lookups, and multi-column data extraction.
' Params:  rowCriteria (Variant)       - Row identifier (row number, range, or search value).
'          srchCol (Variant)           - Column to search in, specified by name or index (required for string search).
'          rtrnCol (Variant, Optional) - Column(s) to return:
'                                      - Single column name or index returns a Range.
'                                      - Array of column names/indices returns a Dictionary.
'                                      - "*" returns all columns in the row as a Dictionary.
'          RaiseSearchError (Boolean, Optional) - Raise error if no match is found (default: True).
' Returns: Range - If retrieving a single value.
'          Dictionary - If retrieving multiple columns or all columns in a row.
' Notes:   - Supports searching by row number, range reference, or string lookup.
'          - Raises ERR_SearchNotFound if no match is found and RaiseSearchError = True.
'          - If rtrnCol is an array or "*", returns a Dictionary of column-value pairs.
' Example: aweSh.SmartCells(25, 1) ' Retrieve Row 25, Col 1
'          aweSh.SmartCells("Mark Watson", "Employee Name") ' Find Mark Watson
'          aweSh.SmartCells("Mary Wilson", "Employee Name", "Email") ' Lookup Email for Mary Wilson
'          aweSh.SmartCells("John", "Name", Array("Age", "City")) ' Returns a Dictionary with multiple columns
'          aweSh.SmartCells("1001", "ID", "*") ' Returns all columns as a Dictionary
'---------------------------------------------------------------------------------------------------
Public Function SmartCells(rowCriteria As Variant, Optional srchCol As Variant = vbNullString, _
                       Optional rtrnCol As Variant = vbNullString, _
                       Optional RaiseSearchError As Boolean = True) As Variant
    On Error GoTo ErrorHandler
    Dim srchColNbr As Long, rowNbr As Long, fndCell As Range, dict As Object, rng As Range, arrIdx As Long
    Set SmartCells = Nothing
    If Not SafeIsEmpty(srchCol) Then srchColNbr = GetColumnNumber(srchCol)

    ' Determine row number from rowCriteria
    Select Case TypeName(rowCriteria)
        Case "Long", "Integer": rowNbr = rowCriteria
        Case "Range":  If rowCriteria.rows.Count > 1 Then RaiseError State_Parameter, "SmartCells", "rowCriteria must be a single row"
                       Set fndCell = rowCriteria.Cells(1, 1): rowNbr = fndCell.Row
                       If srchColNbr = 0 Then srchColNbr = fndCell.Column
        Case "String": Dim srchRng As Range: Set srchRng = ColumnsX(srchCol)
                       Set fndCell = srchRng.Find(What:=rowCriteria, LookIn:=xlValues, LookAt:=xlPart, _
                                                  After:=srchRng.Cells(srchRng.rows.Count, 1))
                       If fndCell Is Nothing Then If RaiseSearchError Then RaiseError State_Search, "SmartCells": Exit Function
                       rowNbr = fndCell.Row
        Case Else:     RaiseError State_Parameter, "SmartCells", "Unsupported rowCriteria type: " & TypeName(rowCriteria)
    End Select

    With m_Worksheet
        ' Handle default empty column request
        If SafeIsEmpty(rtrnCol) Then Set SmartCells = .Cells(rowNbr, srchColNbr): Exit Function
        
        If SafeIsArray(rtrnCol) Then
            If Not IsArray(rtrnCol) Or LBound(rtrnCol) > UBound(rtrnCol) Then Exit Function ' Safe check
            Set dict = CreateObject("Scripting.Dictionary")
            ' Ensure srchCol is included in results if specified
            If Not SafeIsEmpty(srchCol) Then Set dict(ColumnNames(srchColNbr - 1)) = .Cells(rowNbr, srchColNbr)
            For arrIdx = LBound(rtrnCol) To UBound(rtrnCol)
                Set dict(rtrnCol(arrIdx)) = .Cells(rowNbr, GetColumnNumber(rtrnCol(arrIdx))) ' rtnCol Array
            Next arrIdx
            Set SmartCells = dict: Exit Function
        End If
        
        If rtrnCol <> "*" Then Set SmartCells = .Cells(rowNbr, GetColumnNumber(rtrnCol)): Exit Function
        
        ' Wildcard(*) override - All Columns (returns Dictionary)
        Set dict = CreateObject("Scripting.Dictionary")
        For arrIdx = 1 To m_ColumnHdrDict.Count
            Set dict(m_ColumnHdrDict.keys()(arrIdx - 1)) = .Cells(rowNbr, arrIdx)
        Next arrIdx
        Set SmartCells = dict: Exit Function
    End With
    Exit Function

ErrorHandler:
    Set SmartCells = Nothing
    If Err.Number = State_Search Then RaiseError Err.Number, "SmartCells", "Search String Not Found: '" & rowCriteria & "'"
    RaiseError Err.Number, "SmartCells", "Error in SmartCells"
End Function

'###################################################################################################
' Helper Functions - Modular private functions used in this class
'###################################################################################################
'---------------------------------------------------------------------------------------------------
' IsInArray - Checks if a value exists in an array.
' Params:  srchStr (String) - The value to search for.
'          arr (Variant) - The array to search within.
' Returns: Boolean - True if the value exists in the array, otherwise False.
'---------------------------------------------------------------------------------------------------
Public Function IsInArray(srchStr As Variant, arr As Variant) As Boolean
    On Error Resume Next: IsInArray = (Application.match(srchStr, arr, 0) > 0): On Error GoTo 0
End Function

'---------------------------------------------------------------------------------------------------
' IsWorkbookOpen - Checks if a workbook is already open in Excel
' Params:  sFullFileName (String) - The full file name of the workbook to check.
'          wb (Workbook, Optional) - Output workbook object if the file is open.
' Returns: Boolean - True if the workbook is open, False otherwise.
'---------------------------------------------------------------------------------------------------
Public Function IsWorkbookOpen(sFullFileName As String, Optional ByRef wb As Workbook) As Boolean
    Dim sWbNm As String: sWbNm = LCase$(Mid$(sFullFileName, InStrRev(sFullFileName, "\") + 1))
    For Each wb In Application.Workbooks
        If LCase$(wb.Name) = sWbNm Then IsWorkbookOpen = True: Exit Function
    Next
    Set wb = Nothing
End Function

'---------------------------------------------------------------------------------------------------
' ResolveWorksheet - Resolves the specified worksheet by name, object, or table.
' Params:  vTarget (Variant) - Worksheet name or object to resolve.
' Returns: Worksheet - The resolved worksheet object, or raises an error if invalid.
'---------------------------------------------------------------------------------------------------
Private Function ResolveWorksheet(vTarget As Variant) As Worksheet
    Dim ws As Worksheet: Set ws = Nothing
    On Error GoTo ErrorHandler
    Select Case TypeName(vTarget)
        Case "Worksheet": Set ws = vTarget ' Worksheet is already open
        Case "String": On Error Resume Next: Set ws = m_Workbook.Sheets(CStr(vTarget)): On Error GoTo ErrorHandler
    End Select
    Set ResolveWorksheet = ws
    Exit Function
ErrorHandler: RaiseError Err.Number, "ResolveWorksheet", "Error retrieving worksheet."
End Function

' SafeIsArray - Returns True if the variable is an array, suppressing errors.
Public Function SafeIsArray(arr As Variant)
    On Error Resume Next: SafeIsArray = IsArray(arr): On Error GoTo 0
End Function

' SafeIsEmpty - Returns True if a variable is empty, an uninitialized array, or Nothing.
Public Function SafeIsEmpty(v As Variant) As Boolean
    On Error Resume Next: Select Case VarType(v)
        Case vbArray: SafeIsEmpty = (LBound(v) > UBound(v))
        Case vbObject: SafeIsEmpty = (v Is Nothing)
        Case Else: SafeIsEmpty = (IsEmpty(v) Or v = vbNullString)
    End Select: On Error GoTo 0
End Function
